;+
; NAME:
;   MGH_KERNEL
;
; PURPOSE:
;   This function returns one of several pre-defined kernels for
;   filtering operations.
;
; CATEGORY:
;   Time series analysis.
;
; CALLING SEQUENCE:
;   Result = MGH_KERNEL(p0, p1)
;
; POSITIONAL PARAMETERS:
;   p0 (input, string scalar):
;     Filter type. For valid values see code.
;
;   p1 (input, integer or string scalar):
;     An index specifying which kernel of a specific type is to be
;     selected. Valid data types & values depend on p0: see code.
;
; RETURN VALUE:
;   The function returns an array containing the kernel values,
;   normalized to 1
;
; MODIFICATION HISTORY:
;   Mark Hadfield, 2007-10:
;       Written, based on PREKERNEL.
;-
function MGH_KERNEL, p0, p1

   compile_opt DEFINT32
   compile_opt STRICTARR
   compile_opt STRICTARRSUBS
   compile_opt LOGICAL_PREDICATE

   case strlowcase(p0) of

      'hourly':begin            ; Filters for calculating hourly values, Pugh 1987 p 419

         case p1 of

            10: begin
               result = [-0.0000576, $
                         -0.0001725, $
                         -0.0001939, $
                         0.0000589, $
                         0.0007136, $
                         0.0017660, $
                         0.0030310, $
                         0.0041458, $
                         0.0046296, $
                         0.0039967, $
                         0.0018986, $
                         -0.0017344, $
                         -0.0065884, $
                         -0.0119285, $
                         -0.0166494, $
                         -0.0194147, $
                         -0.0188692, $
                         -0.0138889, $
                         -0.0038254, $
                         0.0113057, $
                         0.0307322, $
                         0.0529545, $
                         0.0759010, $
                         0.0971907, $
                         0.1144630, $
                         0.1257208, $
                         0.1296296 ]
            end

            20: begin
               result = [0.0001827, $
                         -0.0002033, $
                         -0.0013622, $
                         -0.0034779, $
                         -0.0035283, $
                         -0.0000000, $
                         0.0081122, $
                         0.0146697, $
                         0.0138889, $
                         -0.0008876, $
                         -0.0235957, $
                         -0.0416666, $
                         -0.0366506, $
                         0.0019899, $
                         0.0708065, $
                         0.1545442, $
                         0.2221317, $
                         0.2500001 ]
            end

            30: begin
               result = [0.0002740, $
                         -0.0020433, $
                         -0.0061020, $
                         0.0000000, $
                         0.0179817, $
                         0.0208333, $
                         -0.0164860, $
                         -0.0625000, $
                         -0.0355647, $
                         0.1062099, $
                         0.2898969, $
                         0.3750000 ]
            end

         endcase

      end

      'low-pass': begin         ; Low-pass filters

         case p1 of

            72: begin
               result = [-0.0000033, $
                         -0.0000126, $
                         -0.0000267, $
                         -0.0000434, $
                         -0.0000598, $
                         -0.0000727, $
                         -0.0000782, $
                         -0.0000727, $
                         -0.0000526, $
                         0.0000145, $
                         0.0000440, $
                         0.0001247, $
                         0.0002281, $
                         0.0003538, $
                         0.0004996, $
                         0.0006622, $
                         0.0008368, $
                         0.0010170, $
                         0.0011954, $
                         0.0013633, $
                         0.0015114, $
                         0.0016297, $
                         0.0017079, $
                         0.0017361, $
                         0.0017049, $
                         0.0016059, $
                         0.0014322, $
                         0.0011787, $
                         0.0008425, $
                         0.0004234, $
                         -0.0000761, $
                         -0.0006504, $
                         -0.0012907, $
                         -0.0019851, $
                         -0.0027183, $
                         -0.0034722, $
                         -0.0042260, $
                         -0.0049562, $
                         -0.0056376, $
                         -0.0062435, $
                         -0.0067464, $
                         -0.0071186, $
                         -0.0073327, $
                         -0.0073629, $
                         -0.0071852, $
                         -0.0067781, $
                         -0.0061239, $
                         -0.0052083, $
                         -0.0040222, $
                         -0.0025611, $
                         -0.0008260, $
                         0.0011764, $
                         0.0034336, $
                         0.0059275, $
                         0.0086342, $
                         0.0115246, $
                         0.0145646, $
                         0.0177159, $
                         0.0209364, $
                         0.0241809, $
                         0.0274025, $
                         0.0305528, $
                         0.0335833, $
                         0.0364465, $
                         0.0390965, $
                         0.0414900, $
                         0.0435877, $
                         0.0453546, $
                         0.0467609, $
                         0.0477829, $
                         0.0484032, $
                         0.0486111 ]
            end

         endcase

      end

      'thompson': begin         ; Thompson (1983)

         case strlowcase(p1) of

            '24m214': begin
               result = [0.003593222, $
                         0.003807495, $
                         0.004275254, $
                         0.005092625, $
                         0.006285915, $
                         0.007814717, $
                         0.009590841, $
                         0.011506394, $
                         0.013462606, $
                         0.015391635, $
                         0.017266285, $
                         0.019096433, $
                         0.020914804, $
                         0.022757428, $
                         0.024645131, $
                         0.026571414, $
                         0.028499651, $
                         0.030369500, $
                         0.032109625, $
                         0.033652232, $
                         0.034944829, $
                         0.035955979, $
                         0.036674104, $
                         0.037100830, $
                         0.037242099  ]
            end

            '24G113': begin
               result = [0.004442175, $
                         0.004477182, $
                         0.004805478, $
                         0.005646540, $
                         0.007030225, $
                         0.008702299, $
                         0.010711246, $
                         0.012951033, $
                         0.014958699, $
                         0.016667614, $
                         0.018168216, $
                         0.019378530, $
                         0.020563951, $
                         0.021992157, $
                         0.023537502, $
                         0.025256197, $
                         0.027200260, $
                         0.029098489, $
                         0.030960144, $
                         0.032836058, $
                         0.034353150, $
                         0.035400978, $
                         0.036137174, $
                         0.036470062, $
                         0.036509277  ]
            end

            '51G113': begin
               result = [-0.000001958, $
                         0.000003870, $
                         0.000012313, $
                         0.000024851, $
                         0.000042627, $
                         0.000067288, $
                         0.000101526, $
                         0.000147761, $
                         0.000208059, $
                         0.000284882, $
                         0.000380791, $
                         0.000498479, $
                         0.000640828, $
                         0.000810529, $
                         0.001010738, $
                         0.001245148, $
                         0.001516670, $
                         0.001827879, $
                         0.002182090, $
                         0.002582394, $
                         0.003031134, $
                         0.003530538, $
                         0.004082390, $
                         0.004687738, $
                         0.005346935, $
                         0.006059150, $
                         0.006822939, $
                         0.007636864, $
                         0.008498320, $
                         0.009403406, $
                         0.010348177, $
                         0.011328058, $
                         0.012337032, $
                         0.013368376, $
                         0.014414844, $
                         0.015468605, $
                         0.016521585, $
                         0.017565049, $
                         0.018589858, $
                         0.019587360, $
                         0.020548516, $
                         0.021463396, $
                         0.022322686, $
                         0.023117922, $
                         0.023840722, $
                         0.024483387, $
                         0.025039259, $
                         0.025502503, $
                         0.025868366, $
                         0.026132808, $
                         0.026292451, $
                         0.026345725  ]
            end

            '120i913': begin
               r0 = [ -0.000360537, $
                      -0.000212081, $
                      -0.000071792, $
                      0.000049698, $
                      0.000147642, $
                      0.000222875, $
                      0.000280202, $
                      0.000325996, $
                      0.000365685, $
                      0.000401795, $
                      0.000433032, $
                      0.000454605, $
                      0.000459624, $
                      0.000441163, $
                      0.000394386, $
                      0.000318158, $
                      0.000215700, $
                      0.000094135, $
                      -0.000036947, $
                      -0.000167514, $
                      -0.000289156, $
                      -0.000396663, $
                      -0.000488692, $
                      -0.000567392, $
                      -0.000637083, $
                      -0.000702370, $
                      -0.000766179, $
                      -0.000828245, $
                      -0.000884426, $
                      -0.000927041, $
                      -0.000946117, $
                      -0.000931275, $
                      -0.000873771, $
                      -0.000768246, $
                      -0.000613794, $
                      -0.000414169, $
                      -0.000177148, $
                      0.000086704, $
                      0.000365519, $
                      0.000647487, $
                      0.000921769, $
                      0.001178827, $
                      0.001410196, $
                      0.001607903, $
                      0.001763818, $
                      0.001869252, $
                      0.001915023, $
                      0.001892088, $
                      0.001792637 ]
               r1 = [  0.001611461, $
                       0.001347253, $
                       0.001003568, $
                       0.000589202, $
                       0.000117891, $
                       -0.000392583, $
                       -0.000921726, $
                       -0.001447799, $
                       -0.001949182, $
                       -0.002405415, $
                       -0.002797825, $
                       -0.003109790, $
                       -0.003326761, $
                       -0.003436243, $
                       -0.003427916, $
                       -0.003294021, $
                       -0.003030001, $
                       -0.002635339, $
                       -0.002114382, $
                       -0.001476983, $
                       -0.000738753, $
                       0.000079158, $
                       0.000950771, $
                       0.001846365, $
                       0.002733785, $
                       0.003579781, $
                       0.004351217, $
                       0.005016057, $
                       0.005544148, $
                       0.005907861, $
                       0.006082735, $
                       0.006048247, $
                       0.005788778, $
                       0.005294783, $
                       0.004564057, $
                       0.003602933, $
                       0.002427191, $
                       0.001062503, $
                       -0.000455716, $
                       -0.002083090, $
                       -0.003767607, $
                       -0.005451338, $
                       -0.007072349, $
                       -0.008566628, $
                       -0.009869873, $
                       -0.010919123, $
                       -0.011654239, $
                       -0.012019343, $
                       -0.011964332, $
                       -0.011446533, $
                       -0.010432516, $
                       -0.008899966, $
                       -0.006839458, $
                       -0.004255895, $
                       -0.001169390, $
                       0.002384577, $
                       0.006355800, $
                       0.010680746, $
                       0.015284565, $
                       0.020083436, $
                       0.024987093, $
                       0.029901225, $
                       0.034729681, $
                       0.039376438, $
                       0.043747394, $
                       0.047752118, $
                       0.051305694, $
                       0.054330728, $
                       0.056759542, $
                       0.058536421, $
                       0.059619747, $
                       0.059983763 ]
               result = [temporary(r0),temporary(r1)]
            end

         endcase

      end

      else: begin

         message, /inform, 'Invalid kernel type: '+p0
         return, [1]

      end

   endcase

   return, [result,reverse(result[0:n_elements(result)-2])]

end



